# ---------------------------------------------------------------------

# Paper:       	N.M. Tran, M. Osipenko, and W.K. Hrdle

#			"Principal Component Analysis in an Asymmetric Norm"

# ---------------------------------------------------------------------

# Quantlet:    	China_example

# ---------------------------------------------------------------------

# Description: 	This is the application of TopDown, BottomUp and  

#			PrincipalExpectile algorithms to the Chinese weather 

#			data

# ---------------------------------------------------------------------

# Usage:       	depends on function libraries fpca.R and 

#			PrincipalDirection.R

# ---------------------------------------------------------------------

# Inputs:      	temperature.txt

# ---------------------------------------------------------------------

# Output:      	Plots and tables shown in the paper

# ---------------------------------------------------------------------

# Example:     	Estimation is done for the expectile levels 0.05, 

#			0.5, and 0.95.

# ---------------------------------------------------------------------

# Author:      	Ngoc Mai Tran, Maria Osipenko

# ---------------------------------------------------------------------

rm(list = ls())
graphics.off()

####enter main directory containning the functional libraries fpca.R and PrincipalDirection.R below
mainDir<- "H:/Asymmetric norm/Rcodes/PCA_in_an_asymmetric_norm/application_weather"
####optional directory for figures
#figDir <- "H:/Asymmetric norm/talk/figures"

library("fda")
source(file.path(mainDir,"fpca.R"))
source(file.path(mainDir,"principalDirection.R"))

temp<-read.table(file.path(mainDir,"temperature.txt"), header=T)
#-------------------substitute missing observations with a mean of neighbour values
# function to substitute some periods of NAs by the mean of neighbours
subna<-function(x){
	indna<-which(is.na(x)==1)
	if(length(indna)>0){
		x[indna]<-(x[min(indna)-1]+x[max(indna)+1])/2
	}
	return(x)
}
temp<-apply(temp,2,subna)

#------------------ function to remove the season Lambda_it
remseas<-function(x){
	time<-seq(1,length(x)); reg<-cbind(time,sin(2*pi*time/365),cos(2*pi*time/365),sin(pi*time/365),cos(pi*time/365))
	fit<-lm(x~reg)
	return(x-fit$fitted.values)
}
tempdes<-apply(temp[,-1],2,remseas)
# ------------------function to remove the AR(p) process
arcust<-function(x){
	mod<-ar(x,aic=F,order.max=10)
	return(mod$resid)
}
tempres<-apply(tempdes,MARGIN=2,arcust)
#-------------------function to average the residuals daywise
tempav<-function(X,over){#X temp data vector, over - timespan
	Xmat<-matrix(X,over,round(length(X)/365,0))
	Xav<-apply(Xmat,1,mean,na.rm=T)
	return(Xav)
}
tempresan<-apply(tempres,2,tempav,over=365)
#-------------------demean the residuals
man<-apply(tempresan,1,mean)
tempresan<-tempresan-man
#--------------------smooth with Fourier series
daybasis<- create.fourier.basis(c(0, 365), nbasis=23, period=365)#nbasis=9
tempmfd <- smooth.basis(day.5, tempresan,daybasis)$fd
tempsm  <- eval.fd(day.5, tempmfd)
#--------------------Compute TD, BUP PCs, and PEC for expectiles
set.seed(1234)
tau	= list(0.05,0.95,0.5)
alpha	= list(-0.45,0.45,0)

#PEC
print("PEC")
output.pec0<-pec.k(tempsm,nk=2,alpha=-0.43)#set as the starting value to speed up the computation
output.pec<-lapply(alpha,pec.k,Y=tempsm, nk=2,reset.tol=50,lab.ini=-output.pec0[[4]])
#BUP
print("BUP")
output.bup<-lapply(tau,bup,Y=tempsm,k=2,basis.start=as.matrix(output.pec[[1]][[2]][,1]))
#TD
print("TD")
output.td<-lapply(X=tau,FUN=laws.main, Y=tempsm,nc=2,max.iter = 30,max.reset = 50, tol = 1e-10, mu = NA,preB = NA,mu.estimate=TRUE,basis.start=output.bup[[2]][[3]])

#---------------------Project the bases to the span of the PEC(0.05) for visual comparison 
#(utility functions listBasis and CompareBasis are in the functional library PrincipalDirection.R)
basis.pec = lapply(output.pec,listBasis,basis0=output.pec[[1]][[2]],num=2)
basis.bup = lapply(output.bup,listBasis,basis0=output.pec[[1]][[2]],num=3)
basis.td  = lapply(output.td,listBasis,basis0=output.pec[[1]][[2]],num=3)

#-------------------------- Table 5: explained variance----------------------------------
#----------------------------------------------------------------------------------------
ess	= matrix(NA,length(tau),3)
for (k in 1:length(tau)){
tau0	= tau[[k]]
sstot	<-  sum(tempsm^2*(sign(tempsm)*(tau0 - 0.5) + 0.5))
ess[k,1] <- 1-resid(tempsm,output.bup[[k]],tau=tau0,rss=T)/sstot
ess[k,2] <- 1-resid(tempsm,output.td[[k]],tau=tau0,rss=T)/sstot
ess[k,3] <- 1-resid.pec(tempsm,output.pec[[k]],tau=tau0,rss=T)/sstot
}
ess	<- data.frame(round(ess,2))
rownames(ess)	= c('tau=0.05','tau=0.95','tau=0.5')
colnames(ess)	= c('BUP','TD','PEC')
write.table(ess, "rss.txt")
ess

#-------------------------------------Figures 3 and 4-------------------------------------
#-----------------------------------------------------------------------------------------
months	= c("J","F","M","A","M","J","J","A","S","O","N","D") # labels for the months
colh	<- c("gray55","gray33","black") # color definition
line	= 2 #line width

#mean
#pdf(file = file.path(figDir,"meanexp_temp.pdf"),paper="special",height=6,width=6)
plot.vec(tempsm,type="l",lwd=2,ylab="",xlab="",ylim=c(-1,1),color=expgr)
lines(output.bup[[1]][[2]],lwd=line,col=colh[1])
lines(output.bup[[2]][[2]],lwd=line,col=colh[1], lty=2)
lines(output.td[[1]][[2]],lwd=line,col=colh[2])
lines(output.td[[2]][[2]],lwd=line,col=colh[2],lty=2)
lines(apply(resid.pec(tempsm,output.pec[[1]],fitted=T,tau=0.05),1,mean)+apply(resid.pec(tempsm,output.pec[[1]],tau=0.05),1,expectile,alpha=0.05-0.5),lwd=line,col=colh[3])
lines(apply(resid.pec(tempsm,output.pec[[2]],fitted=T,tau=0.95),1,mean)+apply(resid.pec(tempsm,output.pec[[2]],tau=0.95),1,expectile,alpha=0.95-0.5),lwd=line,col=colh[3],lty=2)
axis(1,lwd=1,cex.axis=1.3,at=seq(1,365,length=12),lab=months)
axis(2,lwd=1,cex.axis=1.3,at=seq(-1,1,by=0.5))
legend(2,-0.8,c("BUP","TD","PEC"),lty=1,lwd=line, horiz=TRUE,col=colh,bty="n",cex=1)
legend(2,-0.6,c(expression(paste(tau,"=0.05")),expression(paste(tau,"=0.95"))),lty=c(1,2),lwd=line, horiz=TRUE,bty="n",cex=1)
#dev.off()

#PCs
#pdf(file = file.path(figDir,"1PC_temp.pdf"),paper="special",height=6,width=6)
X11()
par(mfrow=c(1,2))
plot(basis.bup[[1]][,1],type="l",lwd=line,axes=FALSE,ylab="",xlab="",ylim=c(-0.2,0.2),col=colh[1])
lines(-basis.bup[[2]][,1],lwd=line,col=colh[1],lty=2)
lines(basis.td[[1]][,1],lwd=line,col=colh[2])
lines(-basis.td[[2]][,1],lwd=line,col=colh[2],lty=2)
lines(basis.pec[[1]][,1],lwd=line,col=colh[3])
lines(-basis.pec[[2]][,1],lwd=line,col=colh[3],lty=2)
lines(seq(1,365), rep(0,365),lwd=1,lty=2)
axis(1,lwd=1,cex.axis=1.3,at=seq(1,365,length=12),lab=months)
axis(2,lwd=1,cex.axis=1.3,at=seq(-0.20,0.2,by=0.1))
legend(2,-0.15,c("BUP","TD","PEC"),lty=1,lwd=line, horiz=TRUE,col=colh,bty="n",cex=1)
legend(2,-0.12,c(expression(paste(tau,"=0.05")),expression(paste(tau,"=0.95"))),lty=c(1,2),lwd=line, horiz=TRUE,bty="n",cex=1)
#dev.off()

#pdf(file = file.path(figDir,"2PC_temp.pdf"),paper="special",height=6,width=6)
plot(basis.bup[[1]][,2],type="l",lwd=line,axes=FALSE,ylab="",xlab="",ylim=c(-0.2,0.2),col=colh[1])
lines(-basis.bup[[2]][,2],lwd=line,col=colh[1],lty=2)
lines(basis.td[[1]][,2],lwd=line,col=colh[2])
lines(-basis.td[[2]][,2],lwd=line,col=colh[2],lty=2)
lines(basis.pec[[1]][,2],lwd=line,col=colh[3])
lines(-basis.pec[[2]][,2],lwd=line,col=colh[3],lty=2)
lines(seq(1,365), rep(0,365),lwd=1,lty=2)
axis(1,lwd=1,cex.axis=1.3,at=seq(1,365,length=12),lab=months)
axis(2,lwd=1,cex.axis=1.3,at=seq(-0.20,0.2,by=0.1))
legend(2,-0.15,c("BUP","TD","PEC"),lty=1,lwd=line, horiz=TRUE,col=colh,bty="n",cex=1)
legend(2,-0.12,c(expression(paste(tau,"=0.05")),expression(paste(tau,"=0.95"))),lty=c(1,2),lwd=line, horiz=TRUE,bty="n",cex=1)
#dev.off()

#-----------------------------------Figure 5-----------------------------------------------
#------------------------------------------------------------------------------------------
#different expectile levels PEC
X11()
par(mfrow=c(1,2))
#pdf(file = file.path(figDir,"pec_temp.pdf"),paper="special",height=6,width=6)
plot(basis.pec[[1]][,1],lwd=2,col=colh[1],ylim=c(-0.2,0.2),type="l",axes=F,ylab="",xlab="")
lines(basis.pec[[1]][,1],lwd=2,col=colh[1])
lines(-basis.pec[[2]][,1],lty=2,lwd=2,col=colh[1])
lines(basis.pec[[3]][,1],lwd=2,col=colh[3])
lines(seq(1,365), rep(0,365),lwd=1,lty=2)
axis(1,lwd=1,cex.axis=1.3,at=seq(1,365,length=12),lab=months)
axis(2,lwd=1,cex.axis=1.3,at=seq(-0.20,0.2,by=0.1))
legend(1,-0.13,c(expression(paste(tau,"=0.05")),expression(paste(tau,"=0.95")),expression(paste(tau,"=0.5"))),lty=c(1,2,1),lwd=2, horiz=TRUE,col=c(colh[1],colh[1],colh[3]),bty="n",cex=0.9)
#dev.off()

#pdf(file = file.path(figDir,"pec2_temp.pdf"),paper="special",height=6,width=6)
plot(basis.pec[[1]][,2],lwd=2,col=colh[1],ylim=c(-0.2,0.2),type="l",axes=F,ylab="",xlab="")
lines(basis.pec[[1]][,2],lwd=2,col=colh[1])
lines(-basis.pec[[2]][,2],lty=2,lwd=2,col=colh[1])
lines(basis.pec[[3]][,2],lwd=2,col=colh[3])
lines(seq(1,365), rep(0,365),lwd=1,lty=2)
axis(1,lwd=1,cex.axis=1.3,at=seq(1,365,length=12),lab=months)
axis(2,lwd=1,cex.axis=1.3,at=seq(-0.20,0.2,by=0.1))
legend(1,-0.13,c(expression(paste(tau,"=0.05")),expression(paste(tau,"=0.95")),expression(paste(tau,"=0.5"))),lty=c(1,2,1),lwd=2, horiz=TRUE,col=c(colh[1],colh[1],colh[3]),bty="n",cex=0.9)
#dev.off()
