# ---------------------------------------------------------------------

# Paper:       	N.M. Tran, M. Osipenko, and W.K. Hrdle

#			"Principal Component Analysis in an Asymmetric Norm"

# ---------------------------------------------------------------------

# Quantlet:    	algeval.R

# ---------------------------------------------------------------------

# Description: 	Function called by simulation_setup.R to parallelize the simulation
# ---------------------------------------------------------------------

# Usage:       	depends on function libraries fpca2.R, 

#			PrincipalDirection2.R
# ---------------------------------------------------------------------

# Inputs:      	-

# ---------------------------------------------------------------------

# Output:      	see below

# ---------------------------------------------------------------------

# Example:     	-

# ---------------------------------------------------------------------

# Author:      	Maria Osipenko

# ---------------------------------------------------------------------




algeval <- function
###Evaluates computing analoges of principal components for expectiles by Bottom Up and TopDown an PrincipalExpectile algorithms for a collection of n curves evaluated at p points
###Returns vectorized columnwise:  1-(nc*3) estimated principal directions by bup, td and pec, 6-9 constant terms, 10[1:3] average mean squared errors, 10[4:6] computational time, 10[7:9] indicator of convergence, 10[10:12] residual sum of squares
(X, 
###vectorized matrix of dimension (2pn+1). first pn elements contain simulated curves (data), next pn elements contain theoretical expectile curves, last elements contains tau expectile level,scenario, n,p.
tol = 1e-10, nc = 2, #number of principal directions to estimate
max.iter = 30, sim = TRUE # indicate whether it is simulated or real data
){
ff<-function(t){
	f=NULL
	for (i in 1:(floor((nc+1)/2)))
	f=cbind(f,sqrt(2)*sin(2*pi/i*t),sqrt(2)*cos(2*pi/i*t))	
	return(f[,1:nc])
}
tau = X[length(X)-3]; n=X[length(X)-1]; p=X[length(X)]; t=seq(0,1,length=p); scen = X[length(X)-2]
Xd  = matrix(X[1:(n*p)],p,n) #extract data
if (sim==TRUE){
X_e  = matrix(X[(n*p+1):(2*n*p)],p,n) #extract theoretical expectiles
} else {X_e=Xd}

#run bottomUp, stores the output in output.bup
print("running BottomUp")
start<-Sys.time()
output.bup <- tryCatch(bup(Xd, k = nc, tau = tau),error=function(e){print(e);out=list(NA,NA,NA,NA,2); return(out)})
end<-Sys.time()
time.bup<-(end-start)
no.bup<-output.bup[[5]]

#run topDown, stores the output in output.td. Note that we have not normalized the topDown basis.
print("running TopDown")
start<-Sys.time()
output.td <- tryCatch(laws.main(Xd, nc, tau, mu.estimate=TRUE),error=function(e){out=list(NA,NA,NA,NA,2); return(out)})
end<-Sys.time()
time.td<-(end-start)
no.td<-output.td[[5]]

#run PrincipalExpectile
print("running PrincipalExpectile")
start<-Sys.time()
output.pec <- tryCatch(pec.k(Xd, alpha=(tau-0.5),nc,reset.opt="random"),error=function(e){out=list(NA,NA,2); return(out)})
end<-Sys.time()
time.pec<-(end-start)
no.pec<-output.pec[[3]]

f <-output.bup[[3]] #fix the basis
if(sim==TRUE){f=ff(t)}

if (no.bup==1){
#residual sum of squares
rsbup=resid(Xd, output.bup, rss = TRUE,tau=tau)
#normalize basis according the true one 
basis.bup <-compareBasis(f, output.bup[[3]])
#mean
mean.bup <- output.bup[[2]]
#amse
amse.bup <- sum((X_e-resid(Xd,output.bup,fitted=T))^2)/(n*p)
} else {
rsbup=NA
basis.bup=matrix(NA,p,nc)
mean.bup=rep(NA,time=p)
amse.bup=NA
}

if (no.td==1){
#residual sum of squares
rstd=resid(Xd, output.td, rss = TRUE,tau=tau)
#normalize basis according the true one 
basis.td <-compareBasis(f, output.td[[3]])
#mean
mean.td <- output.td[[2]]
#amse
amse.td <- sum((X_e-resid(Xd,output.td,fitted=T))^2)/(n*p)
} else {
rstd=NA
basis.td=matrix(NA,p,nc)
mean.td=rep(NA,time=p)
amse.td=NA
}

if (no.pec!=2){
#residual sum of squares
rspec=resid.pec(Xd, output.pec,tau=tau, rss=T)
#normalize basis according the true one 
basis.pec <-compareBasis(f, output.pec[[2]])
#mean
mean.pec <- resid.pec(Xd,output.pec,tau=tau,mean=TRUE)
#amse
amse.pec <- sum((X_e-resid.pec(Xd, output.pec,tau=tau,fitted=T))^2)/(n*p)
} else {
rspec=NA
basis.pec=matrix(NA,p,nc)
mean.pec=rep(NA,time=p)
amse.pec=NA
}

#save the results
rest = c(amse.bup,amse.td,amse.pec,time.bup,time.td,time.pec,no.bup,no.td,no.pec,rsbup,rstd,rspec,tau,scen,n,p)
results= rbind(matrix(basis.bup,nc*p,1),matrix(basis.td,nc*p,1),matrix(basis.pec,nc*p,1),as.matrix(mean.bup),as.matrix(mean.td),as.matrix(mean.pec),as.matrix(rest))

return(results)
}		
	

